from ast import Dict
from typing import Any, Iterable, Optional

from requests import Response, Session
from requests.utils import default_user_agent

from .base import AbstractClient
from .base import ResponseWrapper as BaseResponse


class ResponseWrapper(BaseResponse):
    raw: Response

    def __init__(self, raw: Response):
        self.raw = raw
        self.status = raw.status_code
        self.headers = raw.headers

    def json(self):
        return self.raw.json()

    @property
    def text(self):
        return self.raw.text


class RequestsSession(AbstractClient):
    session: Session

    def __init__(self, instance: Optional[Session] = None):
        super().__init__()
        self.session = instance or Session()

    def request(
        self,
        method: str,
        url: str,
        *,
        scopes: Optional[Iterable[str]] = None,
        headers: Optional[Dict] = None,
        data: Optional[Dict] = None,
        json: Optional[Any] = None,
    ) -> ResponseWrapper:
        response = self.session.request(
            method=method,
            url=url,
            headers=self.prepare_headers(url, headers, scopes),
            data=data,
            json=json,
        )
        return ResponseWrapper(response)

    @property
    def user_agent(self) -> str:
        return f"{default_user_agent()}"



import os
from datetime import datetime
from time import sleep
from typing import Any, List, Literal, Optional, Tuple, Union, Dict
from uuid import UUID

from .base import AbstractApi
from .exceptions import (
    OTMOrderCancellationError,
    OTMFeasibilityError,
    OTMOrderError,
    OTMParametersError,
)
from .helpers import raw_response_to_bytes, bytes_to_file

MAX_CLOUD_COVER_DEFAULT = 15
MIN_OFF_NADIR_RANGE = [0, 45]
MAX_OFF_NADIR_RANGE = MIN_OFF_NADIR_RANGE
MIN_GSD_RANGE = [3.5, 6.8]
MAX_GSD_RANGE = MIN_GSD_RANGE


class OtmV2(AbstractApi):
    """
    Client interface to the OTM API located at
    https://api.satellitevu.com/otm/v2/docs.
    """

    api_path = "otm/v2"
    scopes = []

    def post_feasibility(
        self,
        *,
        contract_id: Union[UUID, str],
        coordinates: Union[Tuple[float, float], Tuple[float, float, float]],
        date_from: datetime,
        date_to: datetime,
        day_night_mode: Literal["day", "night", "day-night"] = "day-night",
        product: Literal["standard", "assured"] = "standard",
        max_cloud_cover: Optional[int] = MAX_CLOUD_COVER_DEFAULT,
        min_off_nadir: Optional[int] = None,
        max_off_nadir: Optional[int] = None,
        min_gsd: Optional[float] = None,
        max_gsd: Optional[float] = None,
        **kwargs,
    ):
        f"""
        Creates a tasking feasibility request.

        Args:
            contract_id: Associated ID of the Contract under which the feasibility
            request will be performed.

            coordinates: An array of coordinates - (longitude, latitude) or
            (longitude, latitude, altitude).

            date_from: datetime representing the start date of the feasibility request.

            date_to: datetime representing the end date of the feasibility request.

            day_night_mode: String representing the mode of data capture. Allowed
            values are ["day", "night", "day-night"]. Defaults to "day-night".

            product: String representing a tasking option. Selecting "assured"
            allows visibility of all passes within the datetime interval. The
            user must accept all cloud cover risk.

            max_cloud_cover: Optional integer, ranging between [0,100] representing
            the maximum threshold of acceptable cloud coverage. Measured in percent.
            Defaults to {MAX_CLOUD_COVER_DEFAULT}.

            min_off_nadir: Optional integer, ranging between {MIN_OFF_NADIR_RANGE},
            representing the minimum angle from the sensor between nadir and the
            scene center. Measured in decimal degrees. Defaults to None.

            max_off_nadir: Optional integer, ranging between {MAX_OFF_NADIR_RANGE},
            representing the maximum angle from the sensor between nadir and the
            scene center. Measured in decimal degrees. Must be larger than
            min_off_nadir. Defaults to None.

            min_gsd: Optional float representing the minimum ground sample
            distance value. Measured in metres, this value reflects the square
            root of the area of the pixel size projected onto the earth. Defaults
            to None.

            max_gsd: Optional float, ranging between {MAX_GSD_RANGE},
            representing the minimum ground sample distance value. Measured in
            metres, this value reflects the square root of the area of the pixel
            size projected onto the earth. Defaults to None.

            Please note that min/max off nadir and min/max gsd are mutually exclusive.
            You must pick either the off nadir angle or gsd as parameters.

        Kwargs:
            Allows sending additional parameters that are supported by the API but not
            added to this SDK yet.

        Returns:
            A dictionary containing properties of the feasibility request.
        """
        url = self.url(f"{str(contract_id)}/tasking/feasibilities/")

        if product == "standard" and not any(
            [min_gsd, max_gsd, min_off_nadir, max_off_nadir]
        ):
            raise OTMParametersError(
                "One pair of Off Nadir or GSD values must be specified for a "
                "standard priority feasibility request."
            )

        payload = {
            "type": "Feature",
            "geometry": {
                "type": "Point",
                "coordinates": coordinates,
            },
            "properties": {
                "datetime": f"{date_from.isoformat()}/{date_to.isoformat()}",
                "product": product,
                **kwargs,
            },
        }

        if product == "standard":
            payload["properties"].update(
                {
                    "satvu:day_night_mode": day_night_mode,
                    "max_cloud_cover": max_cloud_cover,
                }
            )

            for k, v in {
                "min_off_nadir": min_off_nadir,
                "max_off_nadir": max_off_nadir,
                "min_gsd": min_gsd,
                "max_gsd": max_gsd,
            }.items():
                if v is None:
                    continue
                payload["properties"].update({k: v})

        response = self.make_request(
            method="POST", url=url, json={k: v for k, v in payload.items() if v}
        )

        if response.status != 202:
            raise OTMFeasibilityError(response.status, response.text)

        return response.json()

    def get_feasibility(self, *, contract_id: Union[UUID, str], id: Union[UUID, str]):
        """
        Retrieve the feasibility request with a given ID.

        Args:
            contract_id: Associated ID of the Contract for the given feasibility
            request.

            id: UUID representing the feasibility request e.g.
            "2009466e-cccc-4712-a489-b09aeb772296".

        Returns:
            A dictionary containing properties of the feasibility request.
        """
        response = self.make_request(
            method="GET",
            url=self.url(f"{str(contract_id)}/tasking/feasibilities/{str(id)}"),
        )
        return response.json()

    def get_feasibility_response(
        self, *, contract_id: Union[UUID, str], id: Union[UUID, str]
    ):
        """
        Retrieves the feasibility response for the feasibility request
        with a given id.

        Args:
            contract_id: Associated ID of the Contract for the given feasibility
            request.

            id: UUID representing the feasibility request e.g.
            "2009466e-cccc-4712-a489-b09aeb772296".

        Returns:
            A dictionary containing the feasibility response.
        """
        response = self.make_request(
            method="GET",
            url=self.url(
                f"{str(contract_id)}/tasking/feasibilities/{str(id)}/response"
            ),
        )
        return response.json()

    def list_feasibility_requests(
        self,
        *,
        contract_id: Union[UUID, str],
        per_page: int = 25,
        page_token: Optional[str] = None,
    ):
        """
        Retrieves a list of your feasibility requests.

        Args:
            contract_id: Associated ID of the Contract for which all feasibility
            requests will be listed.

            per_page: Number of results (defaults to 25) to be returned per page.

            page_token: Optional string key used to return specific page of results.
            Defaults to None -> assumes page 0.

        Returns:
            A dictionary containing a list of feasibility requests and their properties.
        """
        query = f"per_page={per_page}"
        if page_token:
            query += f"&token={page_token}"

        response = self.make_request(
            method="GET",
            url=self.url(f"{str(contract_id)}/tasking/feasibilities/?{query}"),
        )
        return response.json()

    def create_order(
        self,
        *,
        contract_id: Union[UUID, str],
        coordinates: Optional[
            Union[Tuple[float, float], Tuple[float, float, float]]
        ] = None,
        date_from: Optional[datetime] = None,
        date_to: Optional[datetime] = None,
        day_night_mode: Literal["day", "night", "day-night"] = "day-night",
        product: Literal["standard", "assured"] = "standard",
        max_cloud_cover: Optional[int] = MAX_CLOUD_COVER_DEFAULT,
        min_off_nadir: Optional[int] = None,
        max_off_nadir: Optional[int] = None,
        min_gsd: Optional[float] = None,
        max_gsd: Optional[float] = None,
        addon_withhold: Optional[str] = None,
        signature: Optional[str] = None,
        **kwargs,
    ):
        f"""
        Creates a tasking order request.

        Args:
            contract_id: Associated ID of the Contract under which the tasking
            order will be submitted.

            coordinates: An array of coordinates - (longitude, latitude) or
            (longitude, latitude, altitude). Only required for orders with
            standard priority. Defaults to None.

            date_from: Datetime representing the start date of the order. Required
            for orders with standard priority. Defaults to None.

            date_to: Datetime representing the end date of the order. Required for
            orders with standard priority. Defaults to None

            day_night_mode: String representing the mode of data capture. Allowed
            values are ["day", "night", "day-night"]. Defaults to "day-night".

            product: String representing a tasking option. Selecting "assured"
            allows visibility of all passes within the datetime interval. The
            user must accept all cloud cover risk.

            signature: String representing a signature token required for orders
            with assured priority. Defaults to None.

            max_cloud_cover: Optional integer, ranging between [0,100] representing
            the maximum threshold of acceptable cloud coverage. Measured in percent.
            Defaults to {MAX_CLOUD_COVER_DEFAULT}.

            min_off_nadir: Optional integer, ranging between {MIN_OFF_NADIR_RANGE},
            representing the minimum angle from the sensor between nadir and the
            scene center. Measured in decimal degrees. Defaults to None.

            max_off_nadir: Optional integer, ranging between {MAX_OFF_NADIR_RANGE},
            representing the maximum angle from the sensor between nadir and the
            scene center. Measured in decimal degrees. Must be larger than
            min_off_nadir. Defaults to None.

            min_gsd: Optional float representing the minimum ground sample
            distance value. Measured in metres, this value reflects the square
            root of the area of the pixel size projected onto the earth. Defaults
            to None.

            max_gsd: Optional float, ranging between {MAX_GSD_RANGE},
            representing the minimum ground sample distance value. Measured in
            metres, this value reflects the square root of the area of the pixel
            size projected onto the earth. Defaults to None.

            Please note that min/max off nadir and min/max gsd are mutually exclusive.
            You must pick either the off nadir angle or gsd as parameters.

            addon_withhold: Optional ISO8601 string describing the duration that an order
            will be withheld from the public catalog. Withhold options are specific
            to the contract. If not specified, the option will be set to the default
            specified in the relevant contract. Example values are: ["0d", "3m", "1y"]

        Kwargs:
            Allows sending additional parameters that are supported by the API but not
            added to this SDK yet.

        Returns:
            A dictionary containing properties of the order created.
        """
        url = self.url(f"{str(contract_id)}/tasking/orders/")

        payload = {"properties": {"product": product}}

        if addon_withhold:
            payload["properties"].update({"addon:withhold": addon_withhold})

        if product == "standard":
            if not "coordinates":
                raise OTMParametersError(
                    "`coordinates` must be specified for a standard priority order"
                )
            if not date_from or not date_to:
                raise OTMParametersError(
                    "`date_to` and `date_from` must be specified for a standard "
                    "priority order"
                )

            if not any([min_gsd, max_gsd, min_off_nadir, max_off_nadir]):
                raise OTMParametersError(
                    "One pair of Off Nadir or GSD values must be specified for a "
                    "standard priority order."
                )

        if product == "assured":
            if not signature:
                raise Exception(
                    "Orders with assured priority must have a signature token."
                )
            payload["properties"].update({"signature": signature})

        else:
            payload["properties"].update(
                {
                    "datetime": f"{date_from.isoformat()}/{date_to.isoformat()}",
                    "satvu:day_night_mode": day_night_mode,
                    "max_cloud_cover": max_cloud_cover,
                }
            )

            payload.update(
                {
                    "type": "Feature",
                    "geometry": {
                        "type": "Point",
                        "coordinates": coordinates,
                    },
                }
            )
            payload["properties"].update(kwargs)

            for k, v in {
                "min_off_nadir": min_off_nadir,
                "max_off_nadir": max_off_nadir,
                "min_gsd": min_gsd,
                "max_gsd": max_gsd,
            }.items():
                if v is None:
                    continue
                payload["properties"].update({k: v})

        response = self.make_request(
            method="POST", url=url, json={k: v for k, v in payload.items() if v}
        )

        if response.status != 201:
            raise OTMOrderError(response.status, response.text)

        return response.json()

    def get_order(self, *, contract_id: Union[UUID, str], order_id: Union[UUID, str]):
        """
        Retrieve the order with a given order_id.

        Args:
            contract_id: Associated ID of the Contract under which the tasking
            order is stored.

            order_id: UUID representing the order id e.g.
            "2009466e-cccc-4712-a489-b09aeb772296".

        Returns:
            A dictionary containing properties of the order.
        """
        response = self.make_request(
            method="GET",
            url=self.url(f"{str(contract_id)}/tasking/orders/{str(order_id)}"),
        )
        return response.json()

    def cancel_order(
        self, *, contract_id: Union[UUID, str], order_id: Union[UUID, str]
    ):
        """
        Cancel an order with a given order_id. Raises OTMOrderCancellationError
        if returned status code is not 204.

        Args:
            contract_id: Associated ID of the Contract under which the tasking
            order is stored.

            order_id: UUID representing the order id e.g.
            "2009466e-cccc-4712-a489-b09aeb772296".

        Returns:
            None
        """
        response = self.make_request(
            method="POST",
            url=self.url(f"{str(contract_id)}/tasking/orders/{str(order_id)}/cancel"),
        )
        if response.status != 204:
            raise OTMOrderCancellationError(response.status, response.text)

    def list_orders(
        self,
        *,
        contract_id: Union[UUID, str],
        per_page: int = 25,
        page_token: Optional[str] = None,
    ):
        """
        Retrieves a list of your orders. The orders are returned sorted by creation
        date, with the most recent orders appearing first.

        Args:
            contract_id: Associated ID of the Contract under which all tasking
            orders will be listed.

            per_page: Number of results (defaults to 25) to be returned per page.

            page_token: Optional string key used to return specific page of results.
            Defaults to None -> assumes page 0.

        Returns:
            A dictionary containing a list of orders and their properties.
        """
        query = f"per_page={per_page}"
        if page_token:
            query += f"&token={page_token}"

        response = self.make_request(
            method="GET", url=self.url(f"{str(contract_id)}/tasking/orders/?{query}")
        )
        return response.json()

    def get_price(
        self,
        *,
        contract_id: Union[UUID, str],
        coordinates: Union[Tuple[float, float], Tuple[float, float, float]],
        date_from: datetime,
        date_to: datetime,
        day_night_mode: Literal["day", "night", "day-night"] = "day-night",
        product: Literal["standard", "assured"] = "standard",
        max_cloud_cover: Optional[int] = None,
        min_off_nadir: Optional[int] = None,
        max_off_nadir: Optional[int] = None,
        min_gsd: Optional[float] = None,
        max_gsd: Optional[float] = None,
        **kwargs,
    ):
        """
        Returns the price for a set of ordering parameters.

        Args:
            contract_id: Associated ID of the Contract for which the pricebook
            will be listed.

            coordinates: An array of coordinates.

            date_from: datetime representing the start date of the order.

            date_to: datetime representing the end date of the order.

            day_night_mode: String representing the mode of data capture. Allowed
            values are ["day", "night", "day-night"]. Defaults to "day-night".

            product: String representing a tasking option. Selecting "assured"
            allows visibility of all passes within the datetime interval. The
            user must accept all cloud cover risk.

            max_cloud_cover: Optional integer, ranging between [0,100] representing
            the maximum threshold of acceptable cloud coverage. Measured in percent.
            Defaults to {MAX_CLOUD_COVER_DEFAULT}.

            min_off_nadir: Optional integer, ranging between {MIN_OFF_NADIR_RANGE},
            representing the minimum angle from the sensor between nadir and the
            scene center. Measured in decimal degrees. Defaults to None.

            max_off_nadir: Optional integer, ranging between {MAX_OFF_NADIR_RANGE},
            representing the maximum angle from the sensor between nadir and the
            scene center. Measured in decimal degrees. Must be larger than
            min_off_nadir. Defaults to None.

            min_gsd: Optional float representing the minimum ground sample
            distance value. Measured in metres, this value reflects the square
            root of the area of the pixel size projected onto the earth. Defaults
            to None.

            max_gsd: Optional float, ranging between {MAX_GSD_RANGE},
            representing the minimum ground sample distance value. Measured in
            metres, this value reflects the square root of the area of the pixel
            size projected onto the earth. Defaults to None.

            Please note that min/max off nadir and min/max gsd are mutually exclusive.
            You must pick either the off nadir angle or gsd as parameters.

        Kwargs:
            Allows sending additional parameters that are supported by the API but not
            added to this SDK yet.

        Returns:
            A dictionary containing keys: price, created_at where the price field
            indicates the price, in tokens, for the set of ordering parameters and
            created_at is the UTC datetime at which the price was calculated.

        """
        url = self.url(f"{str(contract_id)}/tasking/price/")

        if product == "standard" and not any(
            [min_gsd, max_gsd, min_off_nadir, max_off_nadir]
        ):
            raise OTMParametersError(
                "One pair of Off Nadir or GSD values must be specified for a "
                "standard priority order."
            )

        payload = {
            "type": "Feature",
            "geometry": {
                "type": "Point",
                "coordinates": coordinates,
            },
            "properties": {
                "datetime": f"{date_from.isoformat()}/{date_to.isoformat()}",
                "product": product,
                **kwargs,
            },
        }

        if product == "standard":
            payload["properties"].update(
                {
                    "satvu:day_night_mode": day_night_mode,
                    "max_cloud_cover": max_cloud_cover,
                }
            )

            for k, v in {
                "min_off_nadir": min_off_nadir,
                "max_off_nadir": max_off_nadir,
                "min_gsd": min_gsd,
                "max_gsd": max_gsd,
            }.items():
                if v is None:
                    continue
                payload["properties"].update({k: v})

        response = self.make_request(
            method="POST", url=url, json={k: v for k, v in payload.items() if v}
        )
        return response.json()

    def search(
        self,
        contract_id: Union[str, UUID],
        page_token: Optional[str] = None,
        per_page: int = 25,
        collections: Optional[List[str]] = None,
        ids: Optional[List[str]] = None,
        date_range: Optional[str] = None,
        created_at: Optional[str] = None,
        updated_at: Optional[str] = None,
        properties: Optional[dict] = None,
        intersects: Optional[Any] = None,
        sort_by: Optional[List[dict]] = None,
    ):
        """
        Performs a search across all feasibility and orders associated with a
        contract.

        Args:
            contract_id: Associated ID of the Contract under which the search
            request will be performed.

            page_token: Optional string key used to return specific page of results.
            Defaults to None -> assumes page 0.

            per_page: Number of results (defaults to 25) to be returned per page.

            collections: List of strings specifying the collections within which
            to search. Allowed values are:
            [ "feasibility", "feasibility|response|%", "orders", "otm|orders|%" ]

            ids: List of strings specifying the ids within which
            to search.

            date_range: Optional string specifying a closed datetime range within
            which to search e.g. "2023-03-22T12:50:24+01:00/2023-03-29T12:50:24+01:00"
            Assumed to be UTC if timezone offset is not given. This datetime range
            corresponds directly to the datetime property associated with the
            order/feasibility request/response item.

            created_at: Optional string specifying a closed datetime range within
            which to search e.g. "2023-03-22T12:50:24+01:00/2023-03-29T12:50:24+01:00"
            Assumed to be UTC if timezone offset is not given. This datetime range
            describes when the database entries are created.

            updated_at: Optional string specifying a closed datetime range within
            which to search e.g. "2023-03-22T12:50:24+01:00/2023-03-29T12:50:24+01:00"
            Assumed to be UTC if timezone offset is not given. This datetime range
            describes when database entries are updated.

            properties: Optional dictionary specifying the filterable fields and
            value. Fields can be filtered by 'status', 'min_off_nadir',
            'max_off_nadir', 'min_gsd', 'max_gsd'.
            Examples: {"status" : "failed", "max_gsd" : 6.5, "min_off_nadir": 35}

            intersects: Optional dictionary with keys "coordinates" and "geometry"
            type that search results intersect with. Available geometry types include:
            "Point","MultiPoint", "LineString", "MultiLineString", "Polygon",
            "MultiPolygon". For example:
            intersects = {"coordinates":[-1.065151, 51.163899], "type" : "Point"}.

            sort_by: List of parameters specifying the field and direction
            ('asc', 'desc') the results are sorted by. Currently only the 'status'
            field is sortable e.g. [{"field": "status", "direction": "desc"}].
        """
        url = self.url(f"{str(contract_id)}/search/")
        payload = {
            "token": page_token,
            "limit": per_page,
            "collections": collections,
            "ids": ids,
            "datetime": date_range,
            "created_at": created_at,
            "updated_at": updated_at,
            "properties": properties,
            "intersects": intersects,
            "sort_by": sort_by,
        }

        response = self.make_request(
            method="POST", url=url, json={k: v for k, v in payload.items() if v}
        )
        return response.json()

    def _download_request(
        self,
        url: str,
        retry_factor: float,
    ):
        """
        Request download, handling retries.
        """
        while True:
            response = self.make_request(method="GET", url=url)

            if response.status == 202:
                sleep(retry_factor * int(response.headers["Retry-After"]))
            elif response.status == 200:
                break

        return response.json()

    def order_download_url(
        self,
        *,
        contract_id: Union[UUID, str],
        order_id: Union[UUID, str],
        retry_factor: float = 1.0,
    ) -> Dict:
        """
        Finds the download url for a submitted tasking order.

        Args:
            contract_id: String or UUID representing the ID of the Contract
            which the order is associated with.

            order_id: String or UUID representing the order id e.g.
            "2009466e-cccc-4712-a489-b09aeb772296".

            retry_factor: A float that determines how retries will be handled.
            A factor of 0.5 means that only half the time specified by the
            "Retry-After" header will be observed before the download request
            is retried again. Defaults to 1.0.

        Returns:
            A dictionary containing the url which the image can be downloaded from.
        """
        url = self.url(
            f"/{contract_id}/tasking/orders/{order_id}/download?redirect=False"
        )

        return self._download_request(url, retry_factor=retry_factor)

    def download_order(
        self,
        *,
        contract_id: Union[UUID, str],
        order_id: UUID,
        destdir: str,
        retry_factor: float = 1.0,
    ):
        """
        Downloads tasking order.

        Args:
            contract_id: String or UUID representing the ID of the Contract
            which an order is associated with.

            order_id: String or UUID representing the order id e.g.
            "2009466e-cccc-4712-a489-b09aeb772296".

            destdir: A string (file path) representing the directory to which
            the imagery will be downloaded.

            retry_factor: A float that determines how retries will be handled.
            A factor of 0.5 means that only half the time specified by the
            "Retry-After" header will be observed before the download request
            is retried again. Defaults to 1.0.

        Returns:
            A string specifying the path the imagery has been downloaded to.
            All items will be downloaded into one ZIP file.
        """
        order_url = self.order_download_url(
            contract_id=contract_id, order_id=order_id, retry_factor=retry_factor
        )["url"]

        response = self.make_request(method="GET", url=order_url)

        destfile = os.path.join(destdir, f"{order_id}.zip")
        data = raw_response_to_bytes(response)

        return bytes_to_file(data, destfile)



from http.client import HTTPResponse
from json import dumps, loads
from sys import version_info
from typing import Any, Dict, Iterable, Optional
from urllib.error import HTTPError
from urllib.parse import urlencode
from urllib.request import Request, urlopen

from .base import AbstractClient
from .base import ResponseWrapper as BaseResponse


class ResponseWrapper(BaseResponse):
    raw: HTTPResponse

    def __init__(self, raw: HTTPResponse):
        self.raw = raw
        self.status = raw.status
        self.headers = {k: v for k, v in raw.getheaders()}

    def json(self):
        return loads(self.raw.read().decode("utf-8"))

    @property
    def text(self):
        return self.raw.read().decode("utf-8")


class UrllibClient(AbstractClient):
    def request(
        self,
        method: str,
        url: str,
        *,
        scopes: Optional[Iterable[str]] = None,
        headers: Optional[Dict] = None,
        data: Optional[Dict] = None,
        json: Optional[Any] = None,
    ) -> ResponseWrapper:
        headers = self.prepare_headers(url, headers, scopes)
        body = None
        if data:
            body = urlencode(data).encode("utf-8")
            headers["Content-Type"] = "application/x-www-form-urlencoded"
        elif json:
            body = dumps(json).encode("utf-8")
            headers["Content-Type"] = "application/json"

        request = Request(method=method, url=url, data=body, headers=headers)
        try:
            response = urlopen(request)
        except HTTPError as error:
            response = error

        wrapper = ResponseWrapper(raw=response)
        return wrapper

    @property
    def user_agent(self) -> str:
        return f"Python-urllib/{version_info[0]}.{version_info[1]}"



class ContractAccessError(Exception):
    def __init__(self, status_code: int, detail: str) -> None:
        self.message = f"Contracts Access Error - {status_code} : {detail}"
        super().__init__(self.message)


class OrdersAPIError:
    def __init__(self, status_code: int, detail: str) -> None:
        self.message = f"Orders API Error - {status_code} : {detail}"
        super().__init__(self.message)


class OTMAPIError(Exception):
    def __init__(self, status_code: int, detail: str) -> None:
        self.message = f"OTM API Error - {status_code} : {detail}"
        super().__init__(self.message)


class OTMOrderError(OTMAPIError):
    pass


class OTMOrderCancellationError(OTMAPIError):
    pass


class OTMFeasibilityError(OTMAPIError):
    pass


class OTMParametersError(Exception):
    pass


class IDAPIError(Exception):
    def __init__(self, status_code: int, detail: str) -> None:
        self.message = f"ID API Error - {status_code} : {detail}"
        super().__init__(self.message)



from io import BytesIO

from satellitevu.http.base import ResponseWrapper


def raw_response_to_bytes(response: ResponseWrapper) -> BytesIO:
    """
    Converts the raw response data from a request into a bytes object.
    """
    raw_response = response.raw

    if isinstance(raw_response, bytes):
        data = BytesIO(raw_response)
    elif hasattr(raw_response, "read"):
        data = BytesIO(raw_response.read())
    elif hasattr(raw_response, "iter_content"):
        data = BytesIO()
        for chunk in raw_response.iter_content():
            data.write(chunk)
        data.seek(0)
    else:
        raise Exception(
            (
                "Cannot convert response object with raw type"
                f"{type(raw_response)} into byte stream."
            )
        )

    return data


def bytes_to_file(data: BytesIO, destfile: str) -> str:
    """
    Converts bytes into a file object at the specified location.
    """
    with open(destfile, "wb+") as f:
        f.write(data.getbuffer())

    return destfile



import os
from time import sleep
from typing import Dict, List, Union
from uuid import UUID


from .base import AbstractApi
from .exceptions import OrdersAPIError
from .helpers import raw_response_to_bytes, bytes_to_file


class OrdersV2(AbstractApi):
    """
    Client interface to the Orders API located at
    https://api.satellitevu.com/orders/v2/docs.
    """

    api_path = "orders/v2"
    scopes = []

    def get_orders(self, *, contract_id: Union[UUID, str]) -> Dict:
        """
         Retrieve details of all imagery orders.

        Args:
            contract_id: String or UUID representing the ID of the Contract
            which an order is associated with.

        Returns:
            A dictionary containing properties of the order.
        """
        url = self.url(f"/{contract_id}/")
        response = self.make_request(method="GET", url=url)

        if response.status != 200:
            raise OrdersAPIError(response.status, response.text)

        return response.json()

    def get_order_details(
        self, *, contract_id: Union[UUID, str], order_id: Union[UUID, str]
    ) -> Dict:
        """
        Retrieve details of an imagery order.

        Args:
            contract_id: String or UUID representing the ID of the Contract
            which an order is associated with.

            order_id: String or UUID representing the order id e.g.
            "2009466e-cccc-4712-a489-b09aeb772296".

        Returns:
            A dictionary containing properties of the order.
        """
        url = self.url(f"/{contract_id}/{order_id}")
        response = self.make_request(method="GET", url=url)

        if response.status != 200:
            raise OrdersAPIError(response.status, response.text)

        return response.json()

    def submit(self, *, contract_id: Union[UUID, str], item_ids: Union[List[str], str]):
        """
        Submit an imagery order for items present in the SatVu catalog.

        Args:
            contract_id: String or UUID representing the ID of the Contract
            under which an order will be created.

            item_ids: A string or list of strings representing the image
            identifiers. For example: "20221005T214049000_basic_0_TABI" or
            ["20221005T214049000_basic_0_TABI, "20221010T222611000_basic_0_TABI"].

        Returns:
            A dictionary containing keys: id, type, features where the id field
            corresponds to an order id and features map to an array of imagery
            items described with conformity to the STAC specification.

        """
        url = self.url(f"/{contract_id}/")

        if isinstance(item_ids, str):
            item_ids = [item_ids]

        return self.make_request(method="POST", url=url, json={"item_id": item_ids})

    def _download_request(
        self,
        url: str,
        retry_factor: float,
    ):
        """
        Request download, handling retries.
        """
        while True:
            response = self.make_request(method="GET", url=url)

            if response.status == 202:
                sleep(retry_factor * int(response.headers["Retry-After"]))
            elif response.status == 200:
                break

        return response.json()

    def item_download_url(
        self,
        *,
        contract_id: Union[UUID, str],
        order_id: Union[UUID, str],
        item_id: str,
        retry_factor: float = 1.0,
    ) -> Dict:
        """
        Finds the download url for an item in a submitted imagery order.

        Args:
            contract_id: String or UUID representing the ID of the Contract
            which an item in the order is associated with.

            order_id: String or UUID representing the order id e.g.
            "2009466e-cccc-4712-a489-b09aeb772296".

            item_id: A string representing the specific image identifiers e.g.
            "20221010T222611000_basic_0_TABI".

            retry_factor: A float that determines how retries will be handled.
            A factor of 0.5 means that only half the time specified by the
            "Retry-After" header will be observed before the download request
            is retried again. Defaults to 1.0.

        Returns:
            A dictionary containing the url which the image can be downloaded from.
        """
        url = self.url(f"/{contract_id}/{order_id}/{item_id}/download?redirect=False")

        return self._download_request(url, retry_factor=retry_factor)

    def order_download_url(
        self,
        *,
        contract_id: Union[UUID, str],
        order_id: Union[UUID, str],
        retry_factor: float = 1.0,
    ) -> Dict:
        """
        Finds the download url for a submitted imagery order.

        Args:
            contract_id: String or UUID representing the ID of the Contract
            which the order is associated with.

            order_id: String or UUID representing the order id e.g.
            "2009466e-cccc-4712-a489-b09aeb772296".

            retry_factor: A float that determines how retries will be handled.
            A factor of 0.5 means that only half the time specified by the
            "Retry-After" header will be observed before the download request
            is retried again. Defaults to 1.0.

        Returns:
            A dictionary containing the url which the image can be downloaded from.
        """
        url = self.url(f"/{contract_id}/{order_id}/download?redirect=False")

        return self._download_request(url, retry_factor=retry_factor)

    def download_item(
        self,
        *,
        contract_id: Union[UUID, str],
        order_id: UUID,
        item_id: str,
        destdir: str,
        retry_factor: float = 1.0,
    ) -> str:
        """
        Download a submitted imagery order.

        Args:
            contract_id: String or UUID representing the ID of the Contract
            which an item in the order is associated with.

            order_id: String or UUID representing the order id e.g.
            "2009466e-cccc-4712-a489-b09aeb772296".

            item_id: A string representing the specific image identifiers e.g.
            "20221010T222611000_basic_0_TABI".

            destdir: A string (file path) representing the directory to which
            the imagery will be downloaded.

            retry_factor: A float that determines how retries will be handled.
            A factor of 0.5 means that only half the time specified by the
            "Retry-After" header will be observed before the download request
            is retried again. Defaults to 1.0.

        Returns:
            A string specifying the path the imagery has been downloaded to.

        """
        item_url = self.item_download_url(
            contract_id=contract_id,
            order_id=order_id,
            item_id=item_id,
            retry_factor=retry_factor,
        )["url"]

        response = self.make_request(method="GET", url=item_url)

        destfile = os.path.join(destdir, f"{item_id}.zip")
        data = raw_response_to_bytes(response)

        return bytes_to_file(data, destfile)

    def download_order(
        self,
        *,
        contract_id: Union[UUID, str],
        order_id: UUID,
        destdir: str,
        retry_factor: float = 1.0,
    ) -> str:
        """
        Downloads entire imagery order.

        Args:
            contract_id: String or UUID representing the ID of the Contract
            which an order is associated with.

            order_id: String or UUID representing the order id e.g.
            "2009466e-cccc-4712-a489-b09aeb772296".

            destdir: A string (file path) representing the directory to which
            the imagery will be downloaded.

            retry_factor: A float that determines how retries will be handled.
            A factor of 0.5 means that only half the time specified by the
            "Retry-After" header will be observed before the download request
            is retried again. Defaults to 1.0.

        Returns:
            A string specifying the path the imagery has been downloaded to.
            All items will be downloaded into one ZIP file.
        """
        order_url = self.order_download_url(
            contract_id=contract_id, order_id=order_id, retry_factor=retry_factor
        )["url"]

        response = self.make_request(method="GET", url=order_url)

        destfile = os.path.join(destdir, f"{order_id}.zip")
        data = raw_response_to_bytes(response)

        return bytes_to_file(data, destfile)



from typing import Any, Union, Dict
from uuid import UUID

from satellitevu.apis.base import AbstractApi
from satellitevu.apis.exceptions import IDAPIError


class IdV2(AbstractApi):
    """
    Client interface to the ID API located at
    https://api.satellitevu.com/id/v2/docs
    """

    api_path = "id/v2"

    def get_user_details(self):
        """
        Retrieves the details of a user.

        Returns:
            A dictionary containing properties of the user.
        """
        url = self.url("user/details")
        response = self.make_request(method="GET", url=url)
        return response.json()

    def rotate_client_secret(self):
        """
        Generate a new client secret for the M2M client associated with a user.

        Returns:
            A dictionary containing the client ID and new secret of the user.
        """
        url = self.url("client/reset")
        response = self.make_request(method="POST", url=url)
        return response.json()

    def edit_user_settings(self, notifications: Dict[str, Any]):
        """
        Update user settings.
        """
        url = self.url("user/settings")
        response = self.make_request(
            method="PUT", url=url, json={"notifications": notifications}
        )
        if response.status != 200:
            raise IDAPIError(response.status, response.text)
        return response.json()

    def get_credit_balance(self, contract_id: Union[UUID, str]):
        """
        Retrieves the credit balance. This is calculated as the credit limit
        for the contract minus the total credits used.
        """
        url = self.url(f"{str(contract_id)}/wallet/credit")
        response = self.make_request(method="GET", url=url)
        if response.status != 200:
            raise IDAPIError(response.status, response.text)
        return response.json()



from importlib.metadata import version
from json import dumps
from unittest.mock import Mock

from mocket import Mocket, Mocketizer
from mocket.mockhttp import Entry
from pytest import mark, skip

from satellitevu.auth.auth import Auth

from . import ResponseWrapper, UrllibClient


@mark.parametrize("method", ("GET", "POST"))
def test_http_client(http_client_class, method):
    client = http_client_class()

    Entry.single_register(
        method, "http://example.com/", body=dumps({"message": "Hello"})
    )
    with Mocketizer():
        response = client.request(method, "http://example.com/")

    assert isinstance(response, ResponseWrapper)
    assert response.json() == {"message": "Hello"}


def test_http_client_user_agent(http_client_class):
    client = http_client_class()

    Entry.single_register(
        "GET", "http://example.com/", body=dumps({"message": "Hello"})
    )

    with Mocketizer():
        client.request("GET", "http://example.com/")
        requests = Mocket.request_list()

    sv_version = version("satellitevu")
    sv_comment = f"(satellitevu/{sv_version})"
    assert requests[0].headers.get("user-agent").endswith(sv_comment)


def test_http_custom_actor(http_client_class):
    if http_client_class == UrllibClient:
        skip("UrllibClient does not support custom instance")

    instance = Mock()
    client = http_client_class(instance=instance)
    client.request("GET", "http://example.com/")

    assert instance.request.called


@mark.parametrize(
    "url, headers, uses_injected_auth",
    (
        ("http://example.com/", None, False),
        ("http://api.example.com/", None, False),
        ("http://api.example.com/non-authed", None, False),
        ("http://api.example.com/authed/", None, True),
        ("http://api.example.com/authed/subpath", None, True),
        (
            "http://api.example.com/authed/subpath",
            {"authorization": "some-other"},
            False,
        ),
    ),
)
def test_http_set_auth(http_client_class, url, headers, uses_injected_auth):
    auth = Mock(wraps=Auth(client_id="mocked", client_secret="mocked"))
    auth.token.return_value = "mock-token"
    Entry.single_register("GET", url)

    client = http_client_class()
    client.set_auth("http://api.example.com/authed/", auth)

    with Mocketizer():
        client.request("GET", url, headers=headers)
        requests = Mocket.request_list()

    assert len(requests) == 1
    assert (
        requests[0].headers.get("authorization") == "Bearer mock-token"
    ) == uses_injected_auth


@mark.parametrize(
    "data, json, body, content_type",
    (
        ({"foo": "bar"}, None, "foo=bar", "application/x-www-form-urlencoded"),
        (None, {"foo": "bar"}, '{"foo": "bar"}', "application/json"),
        (
            {"bar": "foo"},
            {"foo": "bar"},
            "bar=foo",
            "application/x-www-form-urlencoded",
        ),
    ),
)
def test_payload(http_client_class, data, json, body, content_type):
    client = http_client_class()
    Entry.single_register("POST", "http://api.example.com")

    with Mocketizer():
        client.request("POST", "http://api.example.com", data=data, json=json)
        request = Mocket.last_request()
    assert request.headers["content-type"] == content_type
    assert request.body == body



from json import dumps, loads
from itertools import product
from secrets import token_urlsafe
from unittest.mock import patch
from urllib.parse import urlparse
from uuid import uuid4

from mocket import Mocket, mocketize
from mocket.mockhttp import Entry, Response
from pytest import mark, raises

from satellitevu.apis.exceptions import OTMOrderCancellationError, OTMParametersError

API_PATH_FEASIBILITY = "otm/v2/contract-id/tasking/feasibilities/"
API_PATH_ORDERS = "otm/v2/contract-id/tasking/orders/"


@mocketize(strict_mode=True)
def test_cannot_use_v2_without_contract_id(client):
    error_message = "missing 1 required keyword-only argument: 'contract_id'"
    with raises(TypeError) as exc:
        client.otm_v2.get_order(order_id=uuid4())

    assert error_message in str(exc.value)


@mocketize(strict_mode=True)
@mark.parametrize("product", ("standard", "assured"))
def test_post_feasibility(
    oauth_token_entry,
    client,
    otm_request_parameters,
    otm_response,
    product,
):
    contract_id = otm_request_parameters["contract_id"]
    api_path = API_PATH_FEASIBILITY.replace("contract-id", str(contract_id))

    Entry.single_register(
        "POST",
        client._gateway_url + api_path,
        body=dumps(otm_response),
        status=202,
    )

    response = client.otm_v2.post_feasibility(
        **otm_request_parameters,
        product=product,
    )
    assert isinstance(response, dict)

    requests = Mocket.request_list()
    assert len(requests) == 2

    api_request = requests[-1]
    assert api_request.headers["host"] == urlparse(client._gateway_url).hostname
    assert api_request.path == "/" + api_path
    assert api_request.headers["content-type"] == "application/json"
    assert api_request.headers["authorization"] == oauth_token_entry

    api_request_body = loads(api_request.body)
    assert api_request_body["geometry"] == {
        "type": "Point",
        "coordinates": otm_request_parameters["coordinates"],
    }

    properties_keys = ["max_cloud_cover", "min_off_nadir", "max_off_nadir"]

    if product == "assured":
        for key in properties_keys:
            assert not api_request_body["properties"].get(key)
    else:
        for key in properties_keys:
            assert api_request_body["properties"][key] == otm_request_parameters[key]

    assert (
        api_request_body["properties"]["datetime"]
        == f"{otm_request_parameters['date_from'].isoformat()}/{otm_request_parameters['date_to'].isoformat()}"  # noqa: E501
    )


@mocketize(strict_mode=True)
@mark.parametrize("product", ("standard", "assured"))
def test_post_feasibility_off_nadir_gsd_values(
    oauth_token_entry, client, otm_request_parameters, product, otm_response
):
    contract_id = otm_request_parameters["contract_id"]
    api_path = API_PATH_FEASIBILITY.replace("contract-id", str(contract_id))

    otm_request_parameters["product"] = product
    for param in ["min_off_nadir", "max_off_nadir", "min_gsd", "max_gsd"]:
        otm_request_parameters[param] = None

    if product == "assured":
        Entry.single_register(
            "POST",
            client._gateway_url + api_path,
            body=dumps(otm_response),
            status=202,
        )

        response = client.otm_v2.post_feasibility(
            **otm_request_parameters,
        )
        assert isinstance(response, dict)

    else:
        with raises(OTMParametersError):
            client.otm_v2.post_feasibility(**otm_request_parameters)


@mocketize(strict_mode=True)
def test_get_feasibility(
    oauth_token_entry,
    client,
    otm_request_parameters,
    otm_response,
):
    feasibility_id = uuid4()
    contract_id = otm_request_parameters["contract_id"]
    api_path = API_PATH_FEASIBILITY.replace("contract-id", str(contract_id))

    Entry.single_register(
        "GET",
        client._gateway_url + api_path + str(feasibility_id),
        status=200,
        body=dumps(otm_response),
    )

    response = client.otm_v2.get_feasibility(contract_id=contract_id, id=feasibility_id)

    assert isinstance(response, dict)

    requests = Mocket.request_list()
    assert len(requests) == 2

    api_request = requests[-1]
    assert api_request.headers["host"] == urlparse(client._gateway_url).hostname
    assert api_request.path == "/" + api_path + f"{feasibility_id}"
    assert api_request.headers["authorization"] == oauth_token_entry


@mocketize(strict_mode=True)
def test_get_feasibility_response(
    oauth_token_entry,
    client,
    otm_request_parameters,
    otm_response,
):
    feasibility_id = uuid4()
    contract_id = otm_request_parameters["contract_id"]
    api_path = API_PATH_FEASIBILITY.replace("contract-id", str(contract_id))

    Entry.single_register(
        "GET",
        client._gateway_url + api_path + str(feasibility_id) + "/response",
        status=200,
        body=dumps(otm_response),
    )

    response = client.otm_v2.get_feasibility_response(
        contract_id=contract_id, id=feasibility_id
    )

    assert isinstance(response, dict)

    requests = Mocket.request_list()
    assert len(requests) == 2

    api_request = requests[-1]
    assert api_request.headers["host"] == urlparse(client._gateway_url).hostname
    assert api_request.path == "/" + api_path + f"{feasibility_id}/response"
    assert api_request.headers["authorization"] == oauth_token_entry


@mocketize(strict_mode=True)
@mark.parametrize(
    "per_page",
    (None, 10),
)
def test_list_feasibilities(
    oauth_token_entry,
    client,
    otm_request_parameters,
    otm_response,
    per_page,
):
    contract_id = otm_request_parameters["contract_id"]
    per_page = per_page if not None else 25
    api_path = (
        API_PATH_FEASIBILITY.replace("contract-id", str(contract_id)) + f"?{per_page=}"
    )

    Entry.single_register(
        "GET",
        client._gateway_url + api_path,
        status=200,
        body=dumps(otm_response),
    )

    response = client.otm_v2.list_feasibility_requests(
        contract_id=contract_id, per_page=per_page
    )

    assert isinstance(response, dict)

    requests = Mocket.request_list()
    assert len(requests) == 2

    api_request = requests[-1]
    assert api_request.headers["host"] == urlparse(client._gateway_url).hostname
    assert api_request.path == "/" + api_path
    assert api_request.headers["authorization"] == oauth_token_entry


@mocketize(strict_mode=True)
@mark.parametrize("product", ("standard", "assured"))
def test_post_order(
    oauth_token_entry,
    client,
    otm_request_parameters,
    otm_response,
    product,
):
    contract_id = otm_request_parameters["contract_id"]
    api_path = API_PATH_ORDERS.replace("contract-id", str(contract_id))

    if product == "assured":
        otm_request_parameters["product"] = product
        otm_request_parameters["signature"] = token_urlsafe(16)

    Entry.single_register(
        "POST",
        client._gateway_url + api_path,
        body=dumps(otm_response),
        status=201,
    )

    response = client.otm_v2.create_order(**otm_request_parameters)
    assert isinstance(response, dict)

    requests = Mocket.request_list()
    assert len(requests) == 2

    api_request = requests[-1]
    assert api_request.headers["host"] == urlparse(client._gateway_url).hostname
    assert api_request.path == "/" + api_path
    assert api_request.headers["content-type"] == "application/json"
    assert api_request.headers["authorization"] == oauth_token_entry

    api_request_body = loads(api_request.body)

    properties_keys = ["max_cloud_cover", "min_off_nadir", "max_off_nadir"]

    if product == "assured":
        assert "geometry" not in api_request_body.keys()
        assert "signature" in api_request_body["properties"].keys()
        for key in properties_keys:
            assert not api_request_body["properties"].get(key)
    else:
        assert api_request_body["geometry"] == {
            "type": "Point",
            "coordinates": otm_request_parameters["coordinates"],
        }
        assert "signature" not in api_request_body["properties"].keys()
        for key in properties_keys:
            assert api_request_body["properties"][key] == otm_request_parameters[key]
        assert (
            api_request_body["properties"]["datetime"]
            == f"{otm_request_parameters['date_from'].isoformat()}/{otm_request_parameters['date_to'].isoformat()}"  # noqa: E501
        )


@mocketize(strict_mode=True)
@mark.parametrize("product", ("standard", "assured"))
def test_create_order_off_nadir_gsd_values(
    oauth_token_entry,
    client,
    otm_request_parameters,
    product,
    otm_response,
):
    contract_id = otm_request_parameters["contract_id"]
    api_path = API_PATH_ORDERS.replace("contract-id", str(contract_id))

    otm_request_parameters["product"] = product
    otm_request_parameters["signature"] = token_urlsafe(16)
    for param in ["min_off_nadir", "max_off_nadir", "min_gsd", "max_gsd"]:
        otm_request_parameters[param] = None

    if product == "assured":
        Entry.single_register(
            "POST",
            client._gateway_url + api_path,
            body=dumps(otm_response),
            status=201,
        )

        response = client.otm_v2.create_order(**otm_request_parameters)
        assert isinstance(response, dict)
    else:
        with raises(OTMParametersError):
            client.otm_v2.create_order(**otm_request_parameters)


@mocketize(strict_mode=True)
@mark.parametrize(
    "product, withhold",
    product(
        ("standard", "assured"),
        (
            None,
            "0d",
            "7d",
            "1y",
        ),
    ),
)
def test_create_order_with_addons(
    oauth_token_entry,
    client,
    otm_request_parameters,
    product,
    withhold,
    otm_response,
):
    contract_id = otm_request_parameters["contract_id"]
    api_path = API_PATH_ORDERS.replace("contract-id", str(contract_id))

    otm_request_parameters["product"] = product
    otm_request_parameters["signature"] = token_urlsafe(16)
    otm_request_parameters["addon_withhold"] = withhold

    if product == "assured":
        Entry.single_register(
            "POST",
            client._gateway_url + api_path,
            body=dumps(otm_response),
            status=201,
        )

    Entry.single_register(
        "POST",
        client._gateway_url + api_path,
        body=dumps(otm_response),
        status=201,
    )

    response = client.otm_v2.create_order(**otm_request_parameters)
    assert isinstance(response, dict)

    requests = Mocket.request_list()
    assert len(requests) == 2

    api_request = requests[-1]
    assert api_request.headers["host"] == urlparse(client._gateway_url).hostname
    assert api_request.path == "/" + api_path
    assert api_request.headers["content-type"] == "application/json"
    assert api_request.headers["authorization"] == oauth_token_entry

    api_request_body = loads(api_request.body)
    assert api_request_body["properties"]["product"] == product

    if withhold:
        assert api_request_body["properties"]["addon:withhold"] == withhold
    else:
        assert "addon:withhold" not in api_request_body["properties"].keys()


@mocketize(strict_mode=True)
def test_post_assured_order_without_signature(
    oauth_token_entry,
    client,
    otm_request_parameters,
    otm_response,
):
    otm_request_parameters["product"] = "assured"

    with raises(
        Exception,
        match="Orders with assured priority must have a signature token.",
    ):
        client.otm_v2.create_order(**otm_request_parameters)


@mocketize(strict_mode=True)
@mark.parametrize(
    "per_page",
    (None, 10),
)
def test_list_tasking_orders(
    oauth_token_entry,
    client,
    otm_request_parameters,
    otm_response,
    per_page,
):
    contract_id = otm_request_parameters["contract_id"]
    per_page = per_page if not None else 25
    api_path = (
        API_PATH_ORDERS.replace("contract-id", str(contract_id)) + f"?{per_page=}"
    )

    Entry.single_register(
        "GET",
        client._gateway_url + api_path,
        status=200,
        body=dumps(otm_response),
    )

    response = client.otm_v2.list_orders(contract_id=contract_id, per_page=per_page)

    assert isinstance(response, dict)

    requests = Mocket.request_list()
    assert len(requests) == 2

    api_request = requests[-1]
    assert api_request.headers["host"] == urlparse(client._gateway_url).hostname
    assert api_request.path == "/" + api_path
    assert api_request.headers["authorization"] == oauth_token_entry


@mocketize(strict_mode=True)
def test_get_tasking_order(
    oauth_token_entry,
    client,
    otm_request_parameters,
    otm_response,
):
    order_id = uuid4()
    contract_id = otm_request_parameters["contract_id"]
    api_path = API_PATH_ORDERS.replace("contract-id", str(contract_id))

    Entry.single_register(
        "GET",
        client._gateway_url + api_path + str(order_id),
        status=200,
        body=dumps(otm_response),
    )

    response = client.otm_v2.get_order(contract_id=contract_id, order_id=order_id)

    assert isinstance(response, dict)

    requests = Mocket.request_list()
    assert len(requests) == 2

    api_request = requests[-1]
    assert api_request.headers["host"] == urlparse(client._gateway_url).hostname
    assert api_request.path == "/" + api_path + f"{order_id}"
    assert api_request.headers["authorization"] == oauth_token_entry


@mocketize(strict_mode=True)
@mark.parametrize("status_code", (204, 404, 409))
def test_cancel_tasking_order(
    oauth_token_entry, client, otm_request_parameters, status_code
):
    order_id = uuid4()
    contract_id = otm_request_parameters["contract_id"]
    api_path = API_PATH_ORDERS.replace("contract-id", str(contract_id))

    Entry.single_register(
        "POST",
        client._gateway_url + api_path + str(order_id) + "/cancel",
        status=status_code,
    )

    if status_code == 204:
        response = client.otm_v2.cancel_order(
            contract_id=contract_id, order_id=order_id
        )
        assert not response
    else:
        with raises(OTMOrderCancellationError):
            client.otm_v2.cancel_order(contract_id=contract_id, order_id=order_id)

    requests = Mocket.request_list()
    assert len(requests) == 2

    api_request = requests[-1]
    assert api_request.headers["host"] == urlparse(client._gateway_url).hostname
    assert api_request.path == "/" + api_path + f"{order_id}" + "/cancel"
    assert api_request.headers["authorization"] == oauth_token_entry


@mocketize(strict_mode=True)
def test_post_search(
    oauth_token_entry,
    client,
    otm_request_parameters,
    search_response,
):
    contract_id = otm_request_parameters["contract_id"]
    api_path = "otm/v2/contract-id/search/".replace("contract-id", str(contract_id))

    Entry.single_register(
        "POST",
        client._gateway_url + api_path,
        status=200,
        body=dumps(search_response),
    )

    search_parameters = {
        "per_page": 100,
        "collections": ["feasibility"],
        "date_range": "2023-01-01T00:00:00/2023-01-31T11:59:59",
        "properties": {"min_gsd": 15, "max_gsd": 35, "status": "feasible"},
        "intersects": {},
        "sort_by": [{"field": "min_gsd", "direction": "asc"}],
    }

    response = client.otm_v2.search(contract_id=contract_id, **search_parameters)

    assert isinstance(response, dict)

    requests = Mocket.request_list()
    assert len(requests) == 2

    api_request = requests[-1]
    assert api_request.headers["host"] == urlparse(client._gateway_url).hostname
    assert api_request.path == "/" + api_path
    assert api_request.headers["authorization"] == oauth_token_entry


@mocketize(strict_mode=True)
def test_download_order(
    client,
    oauth_token_entry,
    redirect_response,
):
    contract_id = str(uuid4())
    api_path = API_PATH_ORDERS.replace("contract-id", str(contract_id))
    order_id = "528b0f77-5df1-4ed7-9224-502817170613"
    download_dir = "downloads"

    Entry.register(
        "GET",
        client._gateway_url + f"{api_path}{order_id}/download?redirect=False",
        Response(headers={"Retry-After": "1"}, status=202),
        Response(body=dumps(redirect_response), status=200),
    )
    Entry.single_register("GET", uri=redirect_response["url"])

    with patch("satellitevu.apis.otm.bytes_to_file") as mock_file_dl:
        mock_file_dl.return_value = f"{download_dir}/{order_id}.zip"

        response = client.otm_v2.download_order(
            contract_id=contract_id, order_id=order_id, destdir=download_dir
        )

    requests = Mocket.request_list()

    assert len(requests) == 4

    api_request = requests[1]
    assert api_request.headers["host"] == urlparse(client._gateway_url).hostname
    assert api_request.path == f"/{api_path}{order_id}/download?redirect=False"
    assert api_request.headers["authorization"] == oauth_token_entry

    mock_file_dl.assert_called_once()
    assert response == mock_file_dl()
    assert isinstance(response, str)

    Mocket.assert_fail_if_entries_not_served()



from datetime import datetime, timezone
from json import dumps
from urllib.parse import urlparse
from uuid import uuid4

import pytest
from mocket import Mocket, mocketize
from mocket.mockhttp import Entry
from pytest import mark

from satellitevu.auth.exc import Api401Error, Api403Error

API_PATH = "catalog/v1/contract-id/"


@mocketize(strict_mode=True)
@mark.parametrize(
    ["kwargs", "payload"],
    (
        (
            {},
            {"limit": 10},
        ),
        (
            {"limit": 50},
            {"limit": 50},
        ),
        (
            {
                "date_from": datetime(2022, 9, 10, 0, 0, 0, tzinfo=timezone.utc),
                "date_to": datetime(2022, 10, 10, 0, 0, 0, tzinfo=timezone.utc),
            },
            {
                "limit": 10,
                "datetime": "2022-09-10T00:00:00+00:00/2022-10-10T00:00:00+00:00",
            },
        ),
        (
            {"bbox": [0, 0, 1, 1]},
            {"limit": 10, "bbox": [0, 0, 1, 1]},
        ),
    ),
)
def test_search(
    client,
    oauth_token_entry,
    kwargs,
    payload,
):
    contract_id = str(uuid4())
    api_path = API_PATH.replace("contract-id", str(contract_id))

    Entry.single_register(
        "POST", client._gateway_url + f"{api_path}search", "mock-stac-response"
    )

    response = client.catalog_v1.search(contract_id=contract_id, **kwargs)

    requests = Mocket.request_list()
    assert len(requests) == 2

    api_request = requests[-1]
    assert api_request.headers["host"] == urlparse(client._gateway_url).hostname
    assert api_request.path == f"/{api_path}search"
    assert api_request.headers["content-type"] == "application/json"
    assert api_request.headers["authorization"] == oauth_token_entry
    assert api_request.body == dumps(payload)
    assert response.text == "mock-stac-response"


@mocketize(strict_mode=True)
@mark.parametrize(
    "kwargs, payload, status, exception",
    (
        (
            {},
            {"limit": 10},
            401,
            Api401Error,
        ),
        (
            {"limit": 50},
            {"limit": 50},
            403,
            Api403Error,
        ),
    ),
)
def test_unauthorized_search(
    client,
    oauth_token_entry,
    kwargs,
    payload,
    status,
    exception,
):
    contract_id = str(uuid4())
    api_path = API_PATH.replace("contract-id", str(contract_id))

    Entry.single_register(
        "POST",
        client._gateway_url + f"{api_path}search",
        "mock-stac-response",
        status=status,
    )

    with pytest.raises(exception):
        client.catalog_v1.search(contract_id=contract_id, **kwargs)

    requests = Mocket.request_list()
    assert len(requests) == 2

    api_request = requests[-1]
    assert api_request.headers["host"] == urlparse(client._gateway_url).hostname
    assert api_request.path == f"/{api_path}search"
    assert api_request.headers["content-type"] == "application/json"
    assert api_request.headers["authorization"] == oauth_token_entry
    assert api_request.body == dumps(payload)

    Mocket.assert_fail_if_entries_not_served()



from typing import Union
from uuid import UUID

from satellitevu.auth import Auth
from satellitevu.http import AbstractClient

from .base import AbstractApi
from .exceptions import ContractAccessError


class ContractsV1(AbstractApi):
    """
    Interface for information about the contracts and associated products a
    customer has access to.
    """

    api_path = "policy/v1"

    def __init__(self, client: AbstractClient, base_url: str, auth: Auth):
        super().__init__(client, base_url)
        self._auth = auth

    def get_contracts(self):
        url = self.url("/contracts")
        response = self.make_request(
            method="POST",
            url=url,
            json={"token": self._auth.token()},
        )

        if response.status != 200:
            raise ContractAccessError(response.status, response.text)

        return response.json()["result"]

    def get_contract_pricebook(self, contract_id: Union[UUID, str]):
        url = self.url("/policy/query/products")
        response = self.make_request(
            method="POST",
            url=url,
            json={"token": self._auth.token(), "contract_id": str(contract_id)},
        )

        if response.status != 200:
            raise ContractAccessError(response.status, response.text)

        return response.json()["result"]



import json
import re
from urllib.parse import urlparse
from uuid import uuid4

from pytest import raises
from mocket import Mocket, mocketize
from mocket.mockhttp import Entry

from satellitevu.apis.exceptions import IDAPIError


@mocketize(strict_mode=True)
def test_get_user_details(oauth_token_entry, client):
    api_path = "id/v2/user/details"

    Entry.single_register(
        "GET",
        client._gateway_url + api_path,
        status=200,
        body=json.dumps(
            {
                "user_id": "string",
                "name": "string",
                "email": "string",
                "user_metadata": {},
                "last_login": "string",
            }
        ),
    )

    response = client.id_v2.get_user_details()
    assert isinstance(response, dict)

    requests = Mocket.request_list()
    assert len(requests) == 2

    api_request = requests[-1]
    assert api_request.headers["host"] == urlparse(client._gateway_url).hostname
    assert api_request.path == "/" + api_path
    assert api_request.headers["authorization"] == oauth_token_entry


@mocketize(strict_mode=True)
def test_rotate_client_secret(oauth_token_entry, client):
    api_path = "id/v2/client/reset"

    Entry.single_register(
        "POST",
        client._gateway_url + api_path,
        status=200,
        body=json.dumps(
            {
                "client_id": "client_id",
                "client_secret": "client_secret",  # pragma: allowlist secret
            }
        ),
    )

    response = client.id_v2.rotate_client_secret()
    assert isinstance(response, dict)

    requests = Mocket.request_list()
    assert len(requests) == 2

    api_request = requests[-1]
    assert api_request.headers["host"] == urlparse(client._gateway_url).hostname
    assert api_request.path == "/" + api_path
    assert api_request.headers["authorization"] == oauth_token_entry


@mocketize(strict_mode=True)
def test_edit_user_settings(oauth_token_entry, client):
    api_path = "id/v2/user/settings"

    Entry.single_register(
        "PUT",
        client._gateway_url + api_path,
        status=200,
        body=json.dumps(
            {"notifications": {"tasking": {"order_status": {"email": True}}}}
        ),
    )

    response = client.id_v2.edit_user_settings(
        {"tasking": {"order_status": {"email": True}}}
    )
    assert isinstance(response, dict)

    requests = Mocket.request_list()
    assert len(requests) == 2

    api_request = requests[-1]
    assert api_request.headers["host"] == urlparse(client._gateway_url).hostname
    assert api_request.path == "/" + api_path
    assert api_request.headers["content-type"] == "application/json"
    assert api_request.headers["authorization"] == oauth_token_entry


@mocketize(strict_mode=True)
def test_edit_user_settings_invalid_notifications_payload(oauth_token_entry, client):
    api_path = "id/v2/user/settings"

    Entry.single_register(
        "PUT",
        client._gateway_url + api_path,
        status=422,
        body=json.dumps(
            {"detail": [{"loc": ["string"], "msg": "string", "type": "string"}]}
        ),
    )

    with raises(
        IDAPIError,
        match=re.escape(
            'ID API Error - 422 : {"detail": [{"loc": ["string"], "msg": "string", "type": "string"}]}'
        ),
    ):
        client.id_v2.edit_user_settings(
            {"tasking": {"order_status": {"email": "wrong"}}}
        )

    requests = Mocket.request_list()
    assert len(requests) == 2

    api_request = requests[-1]
    assert api_request.headers["host"] == urlparse(client._gateway_url).hostname
    assert api_request.path == "/" + api_path
    assert api_request.headers["content-type"] == "application/json"
    assert api_request.headers["authorization"] == oauth_token_entry


@mocketize(strict_mode=True)
def test_get_credit_balance(oauth_token_entry, client):
    contract_id = str(uuid4())
    api_path = f"id/v2/{contract_id}/wallet/credit"

    Entry.single_register(
        "GET",
        client._gateway_url + api_path,
        status=200,
        body=json.dumps(
            {
                "currency": "GBP",
                "balance": "100000",
            }
        ),
    )

    response = client.id_v2.get_credit_balance(contract_id)
    assert isinstance(response, dict)

    requests = Mocket.request_list()
    assert len(requests) == 2

    api_request = requests[-1]
    assert api_request.headers["host"] == urlparse(client._gateway_url).hostname
    assert api_request.path == "/" + api_path
    assert api_request.headers["authorization"] == oauth_token_entry



import tempfile
from io import BytesIO
from json import dumps
from unittest.mock import patch
from urllib.parse import urljoin, urlparse
from uuid import uuid4

import pytest
from mocket import Mocket, mocketize
from mocket.mockhttp import Entry, Response
from pytest import mark

from satellitevu.apis.orders import bytes_to_file
from satellitevu.auth.exc import Api401Error, Api403Error

API_PATH = "orders/v2/contract-id/"


@mocketize(strict_mode=True)
@mark.parametrize(
    "item_ids",
    (
        ("20221005T214049000_basic_0_TABI"),
        ("20220923T222227000_basic_0_TABI"),
    ),
)
def test_submit_single_item(client, oauth_token_entry, item_ids):
    payload = dumps({"item_id": [item_ids]})
    contract_id = str(uuid4())
    api_path = API_PATH.replace("contract-id", str(contract_id))

    Entry.single_register(
        "POST",
        client._gateway_url + api_path,
        body=dumps(payload),
        status=201,
    )

    response = client.orders_v2.submit(contract_id=contract_id, item_ids=item_ids)

    requests = Mocket.request_list()
    assert len(requests) == 2

    api_request = requests[-1]
    assert api_request.headers["host"] == urlparse(client._gateway_url).hostname
    assert api_request.path == f"/{api_path}"
    assert api_request.headers["content-type"] == "application/json"
    assert api_request.headers["authorization"] == oauth_token_entry
    assert api_request.body == payload

    assert response.status == 201


@mocketize(strict_mode=True)
@mark.parametrize(
    "item_ids",
    (
        (["20221005T214049000_basic_0_TABI", "20220923T222227000_basic_0_TABI"]),
        (["20220923T222227000_basic_0_TABI"]),
    ),
)
def test_submit_multiple_items(
    client,
    oauth_token_entry,
    item_ids,
):
    payload = dumps({"item_id": item_ids})
    contract_id = str(uuid4())
    api_path = API_PATH.replace("contract-id", str(contract_id))

    Entry.single_register(
        "POST",
        client._gateway_url + api_path,
        body=dumps(payload),
        status=201,
    )

    response = client.orders_v2.submit(contract_id=contract_id, item_ids=item_ids)

    requests = Mocket.request_list()
    assert len(requests) == 2

    api_request = requests[-1]
    assert api_request.headers["host"] == urlparse(client._gateway_url).hostname
    assert api_request.path == f"/{api_path}"
    assert api_request.headers["content-type"] == "application/json"
    assert api_request.headers["authorization"] == oauth_token_entry
    assert api_request.body == payload

    assert response.status == 201


@mocketize(strict_mode=True)
def test_item_download_url(
    client,
    oauth_token_entry,
    redirect_response,
):
    order_id = "uuid"
    item_id = "image"
    contract_id = str(uuid4())
    api_path = API_PATH.replace("contract-id", str(contract_id))

    Entry.single_register(
        "GET",
        client._gateway_url + f"{api_path}{order_id}/{item_id}/download?redirect=False",
        body=dumps(redirect_response),
    )

    response = client.orders_v2.item_download_url(
        contract_id=contract_id, order_id=order_id, item_id=item_id
    )

    requests = Mocket.request_list()

    assert len(requests) == 2
    api_request = requests[-1]
    assert api_request.headers["host"] == urlparse(client._gateway_url).hostname
    assert api_request.path == f"/{api_path}uuid/image/download?redirect=False"
    assert api_request.headers["authorization"] == oauth_token_entry

    assert isinstance(response, dict)
    assert response["url"] == "https://image.test"


@mocketize(strict_mode=True)
@mark.parametrize(
    ["status", "exception"],
    (
        (401, Api401Error),
        (403, Api403Error),
    ),
)
def test_no_access_to_download_if_unauthorized(
    client,
    oauth_token_entry,
    status,
    exception,
    redirect_response,
):
    order_id = "uuid"
    item_id = "image"
    contract_id = str(uuid4())
    api_path = API_PATH.replace("contract-id", str(contract_id))

    Entry.single_register(
        "GET",
        client._gateway_url + f"{api_path}{order_id}/{item_id}/download?redirect=False",
        body=dumps(redirect_response),
        status=status,
    )

    with pytest.raises(exception):
        client.orders_v2.item_download_url(
            contract_id=contract_id, order_id=order_id, item_id=item_id
        )

    requests = Mocket.request_list()

    assert len(requests) == 2
    api_request = requests[-1]
    assert api_request.headers["host"] == urlparse(client._gateway_url).hostname
    assert api_request.path == f"/{api_path}uuid/image/download?redirect=False"
    assert api_request.headers["authorization"] == oauth_token_entry


@mocketize(strict_mode=True)
def test_download_order_item(
    client,
    oauth_token_entry,
    redirect_response,
):
    order_id = "uuid"
    item_id = "image"
    contract_id = str(uuid4())
    api_path = API_PATH.replace("contract-id", str(contract_id))
    download_dir = "downloads"

    Entry.register(
        "GET",
        client._gateway_url + f"{api_path}{order_id}/{item_id}/download?redirect=False",
        Response(headers={"Retry-After": "1"}, status=202),
        Response(body=dumps(redirect_response), status=200),
    )
    Entry.single_register("GET", uri=redirect_response["url"])

    with patch("satellitevu.apis.orders.bytes_to_file") as mock_file_dl:
        mock_file_dl.return_value = f"{download_dir}/image.zip"

        response = client.orders_v2.download_item(
            contract_id=contract_id,
            order_id=order_id,
            item_id=item_id,
            destdir=download_dir,
        )

    requests = Mocket.request_list()

    assert len(requests) == 4

    api_request = requests[2]
    assert api_request.headers["host"] == urlparse(client._gateway_url).hostname
    assert api_request.path == f"/{api_path}uuid/image/download?redirect=False"
    assert api_request.headers["authorization"] == oauth_token_entry

    mock_file_dl.assert_called_once()
    assert response == mock_file_dl()
    assert isinstance(response, str)

    Mocket.assert_fail_if_entries_not_served()


@mocketize(strict_mode=True)
def test_get_order_details(
    client,
    oauth_token_entry,
    order_details_response,
):
    fake_uuid = "528b0f77-5df1-4ed7-9224-502817170613"
    contract_id = str(uuid4())
    api_path = API_PATH.replace("contract-id", str(contract_id))

    Entry.single_register(
        "GET",
        client._gateway_url + f"{api_path}{fake_uuid}",
        body=dumps(order_details_response),
    )

    response = client.orders_v2.get_order_details(
        contract_id=contract_id, order_id=fake_uuid
    )

    requests = Mocket.request_list()
    assert len(requests) == 2

    api_request = requests[-1]
    assert api_request.headers["host"] == urlparse(client._gateway_url).hostname
    assert api_request.path == f"/{api_path}{fake_uuid}"
    assert api_request.headers["authorization"] == oauth_token_entry

    assert isinstance(response, dict)


@mocketize(strict_mode=True)
def test_get_orders(
    client,
    oauth_token_entry,
    order_details_response,
):
    contract_id = str(uuid4())
    api_path = API_PATH.replace("contract-id", str(contract_id))

    Entry.single_register(
        "GET",
        client._gateway_url + api_path,
        body=dumps(order_details_response),
    )

    response = client.orders_v2.get_orders(contract_id=contract_id)

    requests = Mocket.request_list()
    assert len(requests) == 2

    api_request = requests[-1]
    assert api_request.headers["host"] == urlparse(client._gateway_url).hostname
    assert api_request.path == f"/{api_path}"
    assert api_request.headers["authorization"] == oauth_token_entry

    assert isinstance(response, dict)


@mocketize(strict_mode=True)
@mark.parametrize(
    ["status", "exception"],
    (
        (401, Api401Error),
        (403, Api403Error),
    ),
)
def test_cannot_get_order_details_if_unauthorized(
    client,
    oauth_token_entry,
    status,
    exception,
    order_details_response,
):
    contract_id = str(uuid4())
    api_path = API_PATH.replace("contract-id", str(contract_id))

    fake_uuid = "528b0f77-5df1-4ed7-9224-502817170613"

    Entry.single_register(
        "GET",
        client._gateway_url + f"{api_path}{fake_uuid}",
        body=dumps(order_details_response),
        status=status,
    )

    with pytest.raises(exception):
        client.orders_v2.get_order_details(contract_id=contract_id, order_id=fake_uuid)

    requests = Mocket.request_list()
    assert len(requests) == 2

    api_request = requests[-1]
    assert api_request.headers["host"] == urlparse(client._gateway_url).hostname
    assert api_request.path == f"/{api_path}{fake_uuid}"
    assert api_request.headers["authorization"] == oauth_token_entry


@mocketize(strict_mode=True)
def test_download_order(
    client,
    oauth_token_entry,
    redirect_response,
):
    contract_id = str(uuid4())
    api_path = API_PATH.replace("contract-id", str(contract_id))
    order_id = "528b0f77-5df1-4ed7-9224-502817170613"
    download_dir = "downloads"

    Entry.register(
        "GET",
        client._gateway_url + f"{api_path}{order_id}/download?redirect=False",
        Response(headers={"Retry-After": "1"}, status=202),
        Response(body=dumps(redirect_response), status=200),
    )
    Entry.single_register("GET", uri=redirect_response["url"])

    with patch("satellitevu.apis.orders.bytes_to_file") as mock_file_dl:
        mock_file_dl.return_value = f"{download_dir}/{order_id}.zip"

        response = client.orders_v2.download_order(
            contract_id=contract_id, order_id=order_id, destdir=download_dir
        )

    requests = Mocket.request_list()

    assert len(requests) == 4

    api_request = requests[1]
    assert api_request.headers["host"] == urlparse(client._gateway_url).hostname
    assert api_request.path == f"/{api_path}{order_id}/download?redirect=False"
    assert api_request.headers["authorization"] == oauth_token_entry

    mock_file_dl.assert_called_once()
    assert response == mock_file_dl()
    assert isinstance(response, str)

    Mocket.assert_fail_if_entries_not_served()


@mocketize(strict_mode=True)
@mark.parametrize(
    ["status", "exception"],
    (
        (401, Api401Error),
        (403, Api403Error),
    ),
)
def test_download_order_unauthorized(
    client,
    redirect_response,
    status,
    exception,
):
    contract_id = str(uuid4())
    api_path = API_PATH.replace("contract-id", str(contract_id))

    order_id = "528b0f77-5df1-4ed7-9224-502817170613"
    download_dir = "downloads"

    Entry.single_register(
        "POST",
        urljoin(client.auth.auth_url, "oauth/token"),
        body=dumps({"access_token": None}),
    )

    Entry.single_register(
        "GET",
        client._gateway_url + f"{api_path}{order_id}/download?redirect=False",
        body=dumps(redirect_response),
        status=status,
    )

    with pytest.raises(exception):
        client.orders_v2.download_order(
            contract_id=contract_id, order_id=order_id, destdir=download_dir
        )

    requests = Mocket.request_list()

    assert len(requests) == 2

    api_request = requests[1]
    assert api_request.headers["host"] == urlparse(client._gateway_url).hostname
    assert api_request.path == f"/{api_path}{order_id}/download?redirect=False"
    assert api_request.headers["authorization"] == "Bearer None"


def test_bytes_to_file():
    outfile_path = tempfile.mkstemp()[1]
    output = bytes_to_file(BytesIO(b"Hello world"), outfile_path)

    assert isinstance(output, str)
